import json

from .exceptions import InvalidInput


def describe(input_format):
  """
  Returns a JSON-able dictionary that describes the 'input_format' by relying on the '__name__' property of any
  functions
  """

  if input_format is None:
    return None
  if type(input_format) is dict:
    output = {}
    for key in input_format:
      output[key] = describe(input_format[key])
    return output
  if type(input_format) is list:
    output = [ describe(input_format[0]), '…' ]
    return output
  if callable(input_format):
    return input_format.__name__
  return str(input_format)


def with_extra_fields(input_format_dict):
  """
  Use this function to wrap a dictionary 'input_format' specification for use with the 'validate' function if you want
  all the properties of ths input data to come through, not just the ones we are explicitly validating.

  Consider the difference between the following two examples:
  validate({ "a": 1, "b": 2 }, { "a": int }) -> { "a": 1 }
  validate({ "a": 1, "b": 2 }, with_extra_fields({ "a": int })) -> { "a": 1, "b": 2 }

  Note that the wrapped input format is still evaluated:
  validate({ "a": "str" }, with_extra_fields({ "a": int })) -> InvalidInput
  """

  if type(input_format_dict) is not dict:
    raise Exception("Expected dictionary")
  def validate_with_extra_fields(input_data_dict):
    if type(input_data_dict) is not dict:
      raise Exception("Expected dictionary")
    output = {}
    output.update(input_data_dict)
    output.update(validate(input_data_dict, input_format_dict))
    return output
  return validate_with_extra_fields


def validate(input_data, input_format, _key=None):
  """
  Validates / parses the input_data against the input_format and returns a new object. Throws an InvalidInput exception
  if there is a problem with the input

  'input_format' can be a dictionary, a single-element list, or a single-argument callable that returns the parsed data.
  These objects can be nested. For lists, the input_format's first element (which must also be an input_format) will be
  applied to all elements in the input_data list. You can specify optional dictionary keys in the input format by
  enclosing them in square brackets. '_key' is internally used for recursion to report exactly what fields failed and
  should not be set.

  Examples:
  validate(None, None) -> None
  validate(5, int) -> 5
  validate("not_an_int", int) -> InvalidInput
  validate({ "n": 5 }, None) -> None
  validate({ "n": 5 }, { "n": int }) -> { "n": 5 } # Note that this is _not_ the original dictionary
  validate({ "n": "5" }, { "n": int }) -> { "n": 5 } # Note that 5 is now an int (because int(5) -> 5)
  validate({ "n": "not_an_int" }, { "n": int }) -> InvalidInput
  validate({ "a": 1 }, { "a": int, "[b]": int }) -> { "a": 1 } # The lack of the optional "b" parameter is ignored
  validate({ "a": 1, "b": 2 }, { "a": int, "[b]": int }) -> { "a": 1, "b": 2 }
  validate({ "a": 1, "b": "not_an_int" }, { "a": int, "[b]": int }) -> InvalidInput
  validate([ 1, "2" ], [ int ]) -> [ 1, 2 ]
  validate([ 1, "not_an_int" ], [ int ]) -> InvalidInput
  """

  if not input_format:
    return None

  try:

    if type(input_format) is dict:
      if type(input_data) is not dict:
        raise Exception('Expected dictionary')
      output = {}
      for key in input_format:
        if len(key) > 2 and key[0] == '[' and key[-1] == ']':
          optional_key = key[1:-1]
          if optional_key in input_data:
            output[optional_key] = validate(input_data[optional_key], input_format[key], _key=optional_key)
        else:
          output[key] = validate(input_data[key], input_format[key], _key=key)
      return output

    if type(input_format) is list:
      if type(input_data) is not list:
        raise Exception('Expected list')
      input_data_size = len(input_data)
      output = input_data_size * [None]
      for i in range(input_data_size):
        output[i] = validate(input_data[i], input_format[0], _key=("%s[%i]" % (_key or "", i)))
      return output

    if callable(input_format):
      return input_format(input_data)

    if (input_format == input_data):
      return input_data

    raise Exception()

  except InvalidInput:
    raise

  except Exception as e:
    e_cause_msg = e.__class__.__name__
    if str(e):
      e_cause_msg += ": " + str(e)
    key_description = ("Value for key '%s'" % _key) if _key else "Value"
    msg = "%s does not match the following input format '%s' - %s" % (key_description, describe(input_format), e_cause_msg)
    raise InvalidInput(msg)



import jwt
import time

from .exceptions import ExpiredToken, InvalidTokenSignature
from .type_validation_functions import anything


class JWTManager:

  def __init__(self, config):
    self.config = config

  def __call__(self, profile_format=anything):
    def decorator(fn):
      if not getattr(fn, '_is_route', False):
        raise Exception("Can only use a 'jwt' decorator over a 'route' decorator")
      fn._profile_format = profile_format
      return fn
    return decorator

  def encode(self, auth_profile):
    """
    Uses the configured parameters to construct a JWT containing the input auth_profile for use with this
    OPService instance. Returns a serializable JWT that can be passed in through the 'Authorization' header on
    subsequent calls using the format "Bearer JWT_TOKEN"

    Note: this method belongs to the OPService class because it is an integral component to how service calls are
    handled and arguments are passed through to functions decorated with 'route'

    Parameters:
      auth_profile: the dictionary to encode as the JWT - note that some claim names are reserved and have special
        meanings ('iss', 'sub', 'aud', 'exp', 'nbf', 'iat', 'jti'). By default, 'iat' and 'exp' are set automatically,
        but precedence is given to any values you pass in.

    Examples:
    >>> plugin = JWTManager({ 'secret': 'ABCDEFGH', 'exp_s': 3600 })
    >>> plugin.encode({ "user_id": "123", "roles": [ "admin" ] })
    >>> # Override the 'exp' property (this JWT will never be valid)
    >>> plugin.encode({ "user_id": "123", "roles": [ "admin" ], "exp": 0 })
    """
    now = int(time.time())
    to_encode = { 'iat': now, 'exp': now + self.config['exp_s'] }
    to_encode.update(auth_profile)
    token = jwt.encode(to_encode, self.config['secret'])
    return str(token, 'UTF-8')

  # See here for more about available options: https://pyjwt.readthedocs.io/en/latest/usage.html
  def decode(self, token, options={ 'require': [ 'iat', 'exp' ] }):
    try:
      return jwt.decode(token, self.config['secret'], options=options)
    except jwt.ExpiredSignatureError as e:
      raise ExpiredToken() from e
    except Exception as e:
      raise InvalidTokenSignature() from e



from .type_validation_functions import anything, int_range


def paginate(required=True, max_size=None, last=anything):
  '''
  This is the 'paginate' interface. It standardizes the way we handle APIs that return lists of data, which allows
  clients to handle paginated APIs uniformly, potentially hiding messy repeated code. This function serves as shorthand
  for building the appropriate input validation format to use with your route definitions. For example, say we are
  defining the following route:

  @service.route('/list-products', None, { 'category_id': int })
  def list_products(profile, data):
    # Returns a list of products ordered by product_id (an int) - we want to be able to use that product_id to pagiante
    ...

  To add the pagination parameters, you can do the following:

  @service.route('/list-products', None, { 'category_id': int, **paginate(last=int) })
  def list_products(profile, data):
    # This injects properties into your input validateion format, resulting in the following:
    # { 'category_id': int, 'pagination': { 'last': int } }
    # It's still up to you to implement the actual pagination logic correctly, this just standardizes the parameters
    ...

  Params:
  - required: whether or not pagination is explicitly required for this route
  - max_size: this allows the client to control the page size (using the 'pagination' / 'size' parameter), but only up
    to the value specified. Omit 'max_size' to control the page size in your route implementation.
  - last: the input validation format for the 'pagination' / 'last' parameter

  Returns the input validation format to be added at the top level of the target route's input validation format
  '''

  key = 'paginate' if required else '[paginate]'
  value = { '[last]': last }
  if max_size:
    value['size'] = int_range('[1, %s]' % max_size)
  return { key: value }



import json
import os


OP_DEV_CONFIG = json.loads(os.environ.get('OP_DEV_CONFIG', '{}'))

OP_DEV_METADATA = json.loads(os.environ.get('OP_DEV_METADATA', '{}')) or { 'domain': None, 'pkg_id': None, 'commit_id': None, 'config_id': None }



import asyncio
import importlib.util
import inspect
import json
import logging
import os
import sys
import traceback

from collections import defaultdict
from contextlib import contextmanager, suppress
from functools import wraps
from op_service_client import OPJSONEncoder
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import Response
from threading import Thread

from .env import OP_DEV_METADATA
from .error_monitor import OPErrorMonitor
from .exceptions import AuthenticationException, ClientException, InvalidAuthorizationHeader, InvalidInput, ServiceException, ignore_exceptions
from .instance_id import SERVICE_INSTANCE_ID
from .jwt_manager import JWTManager
from .logging_config import get_relative_sys_filename
from .request_monitor import RequestMonitor
from .throttle_monitor import ThrottleMonitor
from .type_validation import describe, validate
from .type_validation_functions import non_empty_str, regex


class OPService:

  def __init__(self, jwt_config=None, alarm=None, throttle=None):

    # Note: I generally don't like the concept of "middleware" as it's less clear where / how this code runs.
    # Generally, I prefer to explicitly wrap "middleware" around the "run" function inside the "route" decorator,
    # except where Starlette's built-ins are super convenient...
    middleware = [ Middleware(CORSMiddleware, allow_origins=[ '*' ], allow_headers=["*"], allow_methods=["*"]) ]

    self.starlette = Starlette(middleware=middleware)

    self._routes = []

    self._is_test_instance = os.environ.get('OP_DEV_SERVICE_TEST', False)

    self.jwt = JWTManager(jwt_config) if jwt_config else None
    self.alarm = alarm
    self._error_monitor = OPErrorMonitor(alarm) if alarm else None
    self._request_monitor = RequestMonitor()
    self._throttle_monitor = ThrottleMonitor(throttle) if throttle else None

    # Every production service has a '/' health-check endpoint
    @self.starlette.route('/', methods=[ 'GET' ])
    async def health(request):
      return Response('')

    # TODO: since these three APIs should only ever be called from the context of the op-dev dashboard, you should add
    # some corresponding authentication

    # Returns the commit / config IDs
    @self.route('/__private/metadata', private=True)
    def metadata():
      return OP_DEV_METADATA

    # Returns documentation about the routes
    @self.route('/__private/explore', private=True)
    def explore():
      return self.build_explore_data()

    if self._is_test_instance:
      # This service was brought up in dev mode (using 'op-dev run')

      # An entry point to run the __main__.py script with the provided arguments
      @self.route('/__private/main', { 'args': [ non_empty_str ] }, private=True)
      async def main(data):
        try:
          sys.argv = [ '__main__', *data['args'] ]
          spec = importlib.util.spec_from_file_location('__main__', '/var/www/app/__main__.py')
          module = importlib.util.module_from_spec(spec)
          sys.modules['__main__'] = module
          # Run the module in its own thread so it has its own async context
          t = Thread(target=spec.loader.exec_module, args=(module,))
          t.start()
          t.join()
        except:
          pass
        # try:
        #   if asyncio.iscoroutinefunction(module.main):
        #     await module.main()
        #   else:
        #     module.main()
        # except:
        #   pass

      if jwt_config:
        # If , then add a private API that allows us to
        # generate custom JWTs
        @self.route('/__private/generate-jwt', dict, private=True)
        def generate_jwt(data):
          return self.jwt.encode(data)


  def build_explore_data(self):
    grouped_routes = defaultdict(list)
    for route in self._routes:
      grouped_routes[route['grouping']].append(route)
    output = []
    for key in sorted(grouped_routes):
      output.append(
        {
          'group': key,
          'routes': [
            {
              'name': i['name'],
              'profile': describe(i['fn']._profile_format),
              'data': describe(i['fn']._data_format),
              'documentation': i['fn'].__doc__
            } for i in grouped_routes[key]
          ]
        }
      )
    return output


  # This will only ever get called if we have the 'jwt' property initialized (someone used it as a decorator)
  def _get_input_profile(self, request, profile_format, request_state):
    headers = request.headers
    auth_header_value = headers.get('Authorization')
    if auth_header_value:
      if not auth_header_value.startswith('Bearer '):
        raise InvalidAuthorizationHeader("Authorization header is malformatted - does not start with 'Bearer '")
      token = auth_header_value[7:]
      profile = self.jwt.decode(token)
    else:
      profile = None
    request_state['profile'] = profile
    try:
      request_state['profile_validated'] = validated = validate(profile, profile_format)
      return validated
    except Exception as e:
      raise AuthenticationException() from e


  async def _get_input_data(self, request, data_format, request_state):
    input_data = None
    if data_format is not None:
      content_type = request.headers.get('content-type', '')
      if content_type == 'application/json':
        body = await request.body()
        if body:
          try:
            input_data = json.loads(body)
          except Exception as e:
            raise ClientException('Error parsing JSON') from e
      elif content_type.startswith('multipart/form-data'):
        try:
          form = await request.form()
          def inject_files(d):
            file_id = len(d) == 1 and d.get('__file__')
            return form[file_id] if file_id else d
          input_data = json.loads(form['_'], object_hook=inject_files)
        except Exception as e:
          raise ClientException('Error parsing form data') from e
      else:
        raise ClientException('Unrecognized content-type')
    request_state['input_data'] = input_data
    try:
      request_state['input_data_validated'] = validated = validate(input_data, data_format)
      return validated
    except Exception as e:
      raise InvalidInput(e)

  def expose(self, *keys):
    def decorator(fn):
      if not getattr(fn, '_is_route', False):
        raise Exception("Can only use the 'expose' decorator over a 'route' decorator")
      allowed_keys = [ 'ip_address', 'user_agent' ]
      for key in (keys or [ None ]):
        if key not in allowed_keys:
          raise Exception('Unrecognized property: %s' % key)
      fn._expose = keys
      return fn
    return decorator

  # Functions with this decorator can still be called directly, though they won't get any input / profile validation
  def route(self, route_name, data_format=None, private=False):
    route_name = validate(route_name, regex('(/[a-zA-Z0-9-_]+)+'))
    for route in self._routes:
      if route['name'] == route_name:
        raise Exception('Duplicate routes: ' + route_name)

    def decorator(fn):
      fn._is_route = True
      fn._data_format = data_format
      fn._profile_format = None
      fn._expose = ()

      if not private:
        calling_module = inspect.getmodule(inspect.currentframe().f_back)
        calling_module_name = calling_module and calling_module.__name__ or None
        self._routes.append({ 'name': route_name, 'fn': fn, 'grouping': calling_module_name })

      @self.starlette.route(route_name, methods=[ 'POST' ])
      async def run(request):
        ip_address = request.headers.get('X-Forwarded-For') or request.client.host
        user_agent = request.headers.get('User-Agent')
        request_state = { 'route': route_name, 'ip': ip_address, 'ua': user_agent }
        self._request_monitor.on_route_start(request_state)

        try:
          if self._throttle_monitor:
            self._throttle_monitor.check(ip_address)

          kwargs = {}
          for key in fn._expose:
            if key == 'ip_address':
              kwargs[key] = ip_address
            elif key == 'user_agent':
              kwargs[key] = user_agent
          if fn._profile_format is not None:
            kwargs['profile'] = self._get_input_profile(request, fn._profile_format, request_state)
          if fn._data_format is not None:
            kwargs['data'] = await self._get_input_data(request, fn._data_format, request_state)
          output = fn(**kwargs)
          if asyncio.iscoroutinefunction(fn):
            output = await output
          request_state['response_data'] = content = json.dumps(output, cls=OPJSONEncoder, separators=(',', ':'))
          request_state['response_code'] = 200
          return Response(content=content, media_type='application/json')

        except Exception as e:
          # Make it harder to run a DOS attack by just repeatedly calling APIs with bad input
          await asyncio.sleep(3)

          request_state['exc_info'] = sys.exc_info()
          request_state['response_code'] = status_code = e.status_code if isinstance(e, ServiceException) else 500
          if status_code == 429:
            if self._throttle_monitor:
              with ignore_exceptions(log_msg='Failed to block IP: %s' % ip_address):
                await self._throttle_monitor.block_ip(ip_address)
            if self._error_monitor:
              with ignore_exceptions(log_msg='Alarm failed'):
                self._error_monitor.on_route_end(request_state)
          elif status_code == 500:
            logging.exception('Uncaught exception on route: ' + route_name)
            if self._error_monitor:
              with ignore_exceptions(log_msg='Alarm failed'):
                self._error_monitor.on_route_end(request_state)
            return Response(content='', status_code=500)
          request_state['response_data'] = content = json.dumps({ 'err': e.__class__.__name__, 'msg': str(e) }, separators=(',', ':'))
          return Response(content=content, status_code=status_code, media_type='application/json')

        finally:
          self._request_monitor.on_route_end(request_state)

      return fn
    return decorator

  @contextmanager
  def report_exceptions(self, msg):
    kwargs = { 'metadata': OP_DEV_METADATA }
    try:
      yield
    except Exception:
      logging.exception(msg)
      if self.alarm:
        try:
          _, exc_value, exc_traceback = sys.exc_info()
          stack_summary = traceback.extract_tb(exc_traceback)
          kwargs['exception'] = { 'msg': repr(exc_value), 'traceback': stack_summary.format() }
          self.alarm.report_issue(msg, **kwargs)
        except:
          logging.exception('Failed to report exception')



import json
import signal
import traceback

from datetime import datetime, timezone
from op_service.logging_config import get_relative_sys_filename
from op_service_client import OPJSONEncoder
from threading import Event, Thread

from .abstract_op_service_monitor import AbstractOPServiceMonitor
from .op_json_encoder import OPJSONEncoderFallback
from .env import OP_DEV_METADATA


class OPErrorMonitor(AbstractOPServiceMonitor):


  def __init__(self, reporter):
    self.reporter = reporter


  def on_route_end(self, request_state):
    if (request_state['response_code'] < 500):
      return

    # Any 500-level error surfaced here should have 'exc_info' attached
    exc_type, exc_value, exc_traceback = request_state['exc_info']
    stack_summary = traceback.extract_tb(exc_traceback)
    cause = stack_summary[-1]
    rel_file = get_relative_sys_filename(cause.filename)

    msg = '%s -> %s - %s %s#%s' % (request_state['route'], request_state['response_code'], exc_type.__name__, rel_file, cause.lineno)
    now = datetime.now(timezone.utc)
    issue = {
      'exception': {
        'msg': repr(exc_value),
        'traceback': stack_summary.format()
      },
      'request': {
        'route': request_state['route'],
        'profile': request_state.get('profile_validated'),
        'input_data': request_state.get('input_data'),
        'response_code': request_state['response_code'],
        'response_data': request_state.get('response_data')
      },
      'metadata': OP_DEV_METADATA,
      'count': {
        'n': 1,
        't_start': now.isoformat(),
        't_end': now.isoformat()
      }
    }

    for key in [ 'profile', 'input_data', 'response_data' ]:
      if issue['request'][key]:
        issue['request'][key] = json.dumps(issue['request'][key], cls=OPJSONEncoderFallback)

    self.reporter.report_issue(msg, **issue)



import json
import logging

from .abstract_op_service_monitor import AbstractOPServiceMonitor
from .op_json_encoder import OPJSONEncoderFallback


# This structure describes the order / priority for printing out request state information - the lists are read from top
# to bottom, and keys that come first in a list will be printed instead of any following keys.
STATE_GROUPS = [
  [ 'profile_validated', 'profile' ],
  [ 'input_data_validated', 'input_data' ]
]


# key -> (header, transform)
STATE_KEY_OUTPUTS = {
  'profile': 'p', # Input profile
  'profile_validated': 'pv', # Input profile (validated)
  'input_data': 'i', # Input data
  'input_data_validated': 'iv' # Input data (validated)
}


class RequestMonitor(AbstractOPServiceMonitor):

  def __init__(self):
    self.logger = logging.getLogger('op-request')

  def on_route_start(self, state):
    self.logger.info(json.dumps({ '_': 0, 'r': state['route'], 'ip': state['ip'], 'ua': state['ua'] }, separators=(',', ':')))

  def on_route_end(self, state):
    msg = { '_': 1, 'r': state['route'], 'ip': state['ip'], 'ua': state['ua'], 'c': state['response_code'] }
    for state_group in STATE_GROUPS:
      for key in state_group:
        if key in state:
          msg[ STATE_KEY_OUTPUTS[key] ] = state[key]
          break
    if 'o' in msg:
      # This is a special case: this output is already serialized
      del msg['o']

    packed = json.dumps(msg, cls=OPJSONEncoderFallback, separators=(',', ':'))
    if 'response_data' in state:
      # The 'response_data' is a special case because it is already serialized (and potentially large), so we can avoid
      # reserializing it if we just inject it directly
      packed = packed[:-1] + ',"o":' + state['response_data'] + '}'
    self.logger.info(packed)



"""
This module contains functions designed for use with the 'input_format' parameter for the 'OPService.route' decorator.

Note that it's likely that this module will be imported using '*', so we're careful to obscure any globals using the
'_' prefix.
"""

import datetime as _datetime
import iso8601 as _iso8601
import json as _json
import re as _re
import uuid as _uuid

from fnmatch import fnmatch as _fnmatch
from starlette.datastructures import UploadFile as _UploadFile

from .type_validation import describe as _describe, validate as _validate


def rename(name):
  """
  A decorator for renaming functions - particularly useful for these input validation functions as the actual function
  name is exposed via the /explore route.
  """
  def decorator(f):
    f.__name__ = name
    return f
  return decorator


def regex(pattern):
  """
  An input validation helper for regular expressions. The regex must match the entire string (the '^' and '$' characters
  are implied).
  
  regex("[a-z]+")("apple") -> "apple"
  regex("[a-z]+")("appleJacksCereal") -> Exception
  """
  compiled_pattern = _re.compile(pattern + '$')
  @rename('regex(%s)' % pattern)
  def validate(value):
    match = compiled_pattern.match(value)
    if not match:
      raise Exception('Input does not match regular expression pattern %s' % pattern)
    return value
  return validate


def one_of(*args):
  """
  An input validation helper for enumerations. Uses the 'in' operator to compare the input to the argument
  list. Note that args should be comprised of primitives.
  
  one_of("apple", "ball")("apple") -> "apple"
  one_of("apple", "ball")("car") -> Exception
  """
  @rename('one_of(%s)' % ('%s' % list(args))[1:-1])
  def validate(value):
    if value not in args:
      raise Exception('Input is not in allowed values')
    return value
  return validate


def domain_name(value):
  match = _re.match('(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}$)', value)
  if not match:
    raise Exception('Input is not a valid domain name')
  return value


def _numeric_type_range(range_text, numeric_type_format):
  """
  Parses the input value as a numeric type (using 'numeric'), and makes sure that the value falls within the given
  range, obeying inclusivity vs. exclusivity as denoted by the brackets vs. parantheses.
  
  numeric_range("[0, 10]")(5) -> 5
  numeric_range("[0, 10]")("5") -> 5
  numeric_range("[0, 10]")(-1) -> Exception
  numeric_range("[0, 10]")(0) -> 0
  numeric_range("(0, 10]")(0) -> Exception
  numeric_range("[0, 10]")("not_a_number") -> Exception
  numeric_range("[0, ∞)")(10) -> 10
  numeric_range("[0, inf)")(10) -> 10
  numeric_range("[0, inf]")("inf") -> inf # This returns a float (python parses 'inf' correctly)
  numeric_range("[-∞, ∞]")(0) -> 0 # This useless range matches all numbers...
  """
  condensed_range_text = range_text.replace(' ', '')
  match = _re.match('^(\[|\()([-+]?(\d*\.\d+|\d+|inf|∞)),([-+]?(\d*\.\d+|\d+|inf|∞))(\]|\))$', condensed_range_text)
  min_inclusive = (match.group(1) == '[')
  from_min = numeric(match.group(2).replace('∞', 'inf'))
  to_max = numeric(match.group(4).replace('∞', 'inf'))
  max_inclusive = (match.group(6) == ']')

  @rename('%s(%s)' % (numeric_type_format.__name__, range_text))
  def validate(value):
    numeric_value = numeric_type_format(value)
    if not (min_inclusive and (numeric_value >= from_min) or (numeric_value > from_min)):
      raise Exception('Input value is too small')
    if not (max_inclusive and (numeric_value <= to_max) or (numeric_value < to_max)):
      raise Exception('Input value is too large')
    return numeric_value
  return validate


def int_range(range_text):
  return _numeric_type_range(range_text, int)


def numeric_range(range_text):
  return _numeric_type_range(range_text, numeric)


def numeric(value):
  """
  Parses this value as a numeric type, first trying 'int' then 'float'
  
  numeric(1.5) -> 1.5
  numeric("1.5") -> 1.0
  numeric(1) -> 1
  numeric(1.0) -> 1.0
  numeric("1.0") -> 1.0
  numeric("not_numeric") -> Exception
  """
  if isinstance(value, float):
    return value
  try:
    return int(value)
  except ValueError:
    return float(value)


def iso_date_str(value):
  if type(value) == _datetime.date:
    return value
  # The iso8601 module is a little too forgiving here, so first check against this regex:
  value = regex('\d{4}-\d{2}-\d{2}')(value)
  d = _iso8601.parse_date(value)
  return d.date()


def iso_datetime_str(value):
  # The input value is required to have a timezone, so this is really RFC 3339 instead of ISO 8601
  if type(value) == _datetime.datetime:
    d = value
  else:
    d = _iso8601.parse_date(value, default_timezone=None)
  if (d.tzinfo is None) or (d.tzinfo.utcoffset(d) is None):
    raise Exception('Missing timezone')
  return d


def non_empty_str(value):
  return truthy(str)(value)


def callable_with_arity(i):
  @rename('callable_with_arity(%i)' % i)
  def validate(value):
    if not callable(value):
      raise Exception('Value is not callable')
    arg_count = value.__code__.co_argcount
    if (arg_count != i):
      raise Exception('Expected callable to take %i arguments (got %i)' % (i, arg_count))
    return value
  return validate


def anything(value):
  return value


def uuid_str(value):
  # Value is either already a UUID or a string that can be parsed as one
  return value if type(value) == _uuid.UUID else _uuid.UUID(value)


@rename('JSON blob')
def json_blob(value):
  _json.loads(value)
  return value


def file_upload(content_types=[], max_size=None):
  conditions = [ content_types and 'content_types=%s' % content_types, max_size and 'max_size=%s' % max_size ]
  conditions_str = (', ').join([ i for i in conditions if i ])
  @rename('file_upload%s' % ('(%s)' % conditions_str if conditions_str else '') )
  def validate(value):
    if not isinstance(value, _UploadFile):
      raise Exception('Expected file')
    if content_types and not any([ _fnmatch(value.content_type, x) for x in content_types ]):
      raise Exception('Unexpected content-type')
    if max_size:
      value.file.seek(0, 2)
      if value.file.tell() > max_size:
        raise Exception('File is too large')
      value.file.seek(0)
    return value
  return validate


# Short for variable_dict
def v_dict(key=anything, value=anything):
  @rename('{ %s: %s, … }' % (_describe(key), _describe(value)))
  def validate(input_value):
    if type(input_value) is not dict:
      raise Exception('Expecting dictionary')
    output = {}
    for i, j in input_value.items():
      output[ _validate(i, key) ] = _validate(j, value)
    return output
  return validate


@rename('true')
def is_true(value):
  if value is True:
    return True
  raise Exception()


# The following functions cannot be used alone, but instead can be used to chain / compose various other input
# validation functions:


def truthy(condition):
  @rename('truthy(%s)' % _describe(condition))
  def validate(value):
    if not value:
      raise Exception('Input is not truthy')
    return _validate(value, condition)
  return validate


def nullable(condition):
  @rename('nullable(%s)' % _describe(condition))
  def validate(value):
    if value is None:
      return value
    return _validate(value, condition)
  return validate


def non_nullable(condition):
  @rename('non_nullable(%s)' % _describe(condition))
  def validate(value):
    if value is None:
      raise Exception('Input cannot be null')
    return _validate(value, condition)
  return validate


# It's not always constructive / informative to include the entire type validation as part of the input format because
# it can be hard to describe. The following functions are accessors for already-validated input that provide additional
# validation beyond what makes sense to describe in the input format.

def only_item(one_item_dict):
  """
  Extracts a single key / value pair as a tuple from the input dictionary

  Typical use would be with an input format that requires only one of a number of optional keys:
  { "[a]": str, "[b]": int, "[c]": { "x": int, "y": int } }

  only_item({ "a": "abc" }) -> ("a", "abc")
  only_item({ "b": 123 }) -> ("b", 123)
  only_item({ "d": { "x": 1, "y": 2 } }) -> ("d", { "x": 1, "y": 2 })
  only_item({ "a": "123" "b": 123 }) -> Exception
  """

  if (len(one_item_dict) != 1):
    raise Exception('Expected exactly one key')
  for item in one_item_dict.items():
    return item



class AbstractOPServiceMonitor(object):
  """
  An abstract base class that can be used to execute custom code upon starting and finishing each request to an
  OPService instance. Add a monitor instance to an OPService instance using the 'add_monitor' function. Note that
  monitors will be evaluated in the same order in which they are added.

  Note that these monitors are only triggered on calls that explicitly follow the OP pattern (see the
  'OPService._is_op_request' function).

  The functions 'start' and 'finish' may be called asynchronously, though you are guaranteed that 'start' will complete
  before 'finish' is called.

  Example - a monitor that prints the route name and time taken for each request (regardless of success):

  import threading
  from datetime import datetime
  from op_service.service import OPService

  class TimerMonitor(AbstractOPServiceMonitor):

    def on_route_start(self, state):
      threading.local().start = datetime.now()

    def on_route_end(self, state):
      start = threading.local().start
      end = datetime.now()
      print("%s: %fs" % (state['route'], (end - start).total_seconds()))

  OP = OPService.instance()
  OP.add_monitor(TimerMonitor())
  """


  def on_route_start(self, state):
    """
    This function will be called immediately as soon as a request is received (even before the request data is
    deserialized or the auth headers have been processed).

    Parameters:
      state: a dictionary that describes the request state. Treat it immutably as it can be shared between threads.
        Keys:
        - metadata: the service metadata object: { 'pkg_id': ..., 'commit_id': ..., 'config_id': ... }
        - route: the name of the route at which this request is directed
    """


  def on_route_end(self, state):
    """
    This function will be called after a request ends. It cannot affect the response. The 'state' dictionary contains
    properties that can be used to determine what response was sent.

    Parameters:
      state: a dictionary that describes the request state. Treat it immutably as it can be shared between threads. Note
        that not all of the keys listed below will be present depending on whether or not the request failed (and at
        what point) - the reason in brackets indicates why the key may be missing.
        Keys:
        - metadata: the service metadata object: { 'pkg_id': ..., 'commit_id': ..., 'config_id': ... }
        - route: the name of the route at which this request is directed
        - response_code: the HTTP integer code of the response (i.e. 200)
        - response_data: the serialized JSON response data
        - exc_info: [only if an exception was thrown] the python exc_info object of the thrown exception
        - profile: [only if the the route specifies a profile validation format] the decoded JWT profile from the call
          headers
        - profile_validated: [only if the the route specifies a profile validation format and the raw profile passes that
          validation] the validated JWT profile (see 'profile' key)
        - input_data: [only if the route specifies an input data validation format] the deserialized input data
        - input_data_validated: [only if the route specifies an input data validation format and the input data passes
          that validation] the validated input data (see 'input_data' key)
    """



import setuptools
import sys

assert sys.version_info >= (3, 6)

setuptools.setup(
  name='op-service',
  version='0.1.2',
  author='Alex Exarhos',
  author_email='AlexExarhos@gmail.com',
  packages=[
    'op_service'
  ],
  install_requires=[
    'iso8601 >= 0.1.11',
    'op-alarm >= 0.1.0',
    'op-distributed-task >= 0.1.0',
    'op-service-client >= 0.1.0',
    'op-storage >= 0.1.0',
    'PyJWT >= 1.4.2',
    'python-multipart >= 0.0.5',
    'starlette >= 0.13.0'
  ]
)



"""
This module provides the 'cache' decorator - an LRU cache that works very similarly to the 'functools.lru_cache', but
with a couple of important differences:
- The 'ttl_s' paramater (time-to-live): allows cached values to expire after the given time, which means that a function
  call after this time forces the function to be re-evaluated. Note that the stale cache values will persist in memory
  until a successive call refreshes them.
- Better support for multi-threading: if two threads simultaneously make the same call with the same parameters, only
  one of the threads will actually do the work of evaluating the function. The other thread(s) will wait on that result
  and return it, or raise any exception that was thrown (though note that exceptions will not be cached).
Arguments to a cached function must be hashable (no dictionaries).

Parameters:
  ttl_s: [optional] that amount of time after which a cached value is no longer valid, or 'None' to cache values forever
    (defaults to 'None')
  max_size: [optional] the max number of items to store in the cache before we start eviction, or 'None' for no limit
    (defaults to 'None')

This cache decorator also attaches a 'cache' property to the decorated function, which has the following properties:
  ttl_s: the ttl_s the cache was created with
  max_size: the max_size the cache was created with
  func: a reference to the original underlying function (calling directly will not impact the cache metrics)
  get: function that takes the same input as the wrapped function but returns the CacheEntry instead of the raw value
  contents: on OrderedDict containing the current contents of the cache
  evict: function that takes the same input as the wrapped function and evicts any cached result
  info: returns a CacheInfo object summarizing the the cache



# <op-test>

from op_service.cache import cache
from threading import Lock, Thread

class Work():
  def __init__(self):
    self.work = []
    self.work_lock = Lock()
  def do(work_id=None):
    self.work_lock.acquire()
    self.work.append(work_id)
    self.work_lock.release()

def try_except_pass(f):
  def run(*args, **kwargs):
    try:
      f(args, kwargs)
    except:
      pass
  return run

# <op-test>
work = Work()
@cache()
def test():
  work.do()
test()
test() # Uses cache - does not do work
assert len(work.work) == 1

# <op-test>
work = Work()
@cache()
def test():
  work.do()
  raise Exception()
try_except_pass(test)()
try_except_pass(test)() # Does not cache exceptions - does do work
assert len(work.work) == 2

# <op-test>
work = Work()
@cache(max_size=1)
def test(i):
  work.do(i)
test(1)
test(1) # Uses cache - does not do work
test(2)
test(1) # The result for this key fell out of the cache - does do work
assert work.work == [1, 2, 1]

# <op-test>
work = Work()
@cache(ttl_s=0.05)
def test():
  work.do(i)
test()
test() # Uses cache - does not do work
sleep(0.1)
test() # Value in cache has expired - does do work
assert len(work.work) == 2

# <op-test>
# This example demonstrates a difference from the "functools.lru_cache" related to multi-threading - even though no
# value has yet been cached by the time the second thread hits the function, it is aware that another thread is already
# working on the result, so it will block and return the result form that other thread.
work = Work()
@cache()
def test():
  work.do()
  sleep(0.1) # To make sure both threads would have time to get in here
Thread(target=test).start()
sleep(0.05)
test() # Waits on the result from the first thread (even though a result is not cached yet) - does not do work
assert len(work.work) == 1

# <op-test>
# The same logic applies to exceptions. Note that this is somewhat akin to caching an exception, but only for threads
# whose calls happened to overlap. Successive calls will not find a cached exception.
work = Work()
@cache()
def test():
  work.do()
  sleep(1)
  raise Exception()
Thread(target=try_except_pass(test)).start()
sleep(0.1)
Thread(target=try_except_pass(test)).start() # Waits on the result from the first thread - does not do work
sleep(1) # Both threads will finish during this statement
test_exception() # The first result came back as an exception, so it was not cached - does do work
assert len(work.work) == 2

# <op-test>
# The special multi-threading redundancy improvement does _not_ apply if we are dealing with the same thread. The
# following code will result in a "maximum recursion depth exceeded" error as opposed to hanging.
try:
  @cache()
  def test():
    test()
except RecursionError:
  pass

# </op-test>



"""


from collections import namedtuple, OrderedDict
from contextlib import suppress
from functools import wraps
from threading import Lock, Timer
from time import time


class CacheKey():

  # 'args' is a tuple and 'kwargs' is a dictionary (where all the keys are strings)
  def __init__(self, args, kwargs):
    self.args = args
    self.kwargs = kwargs
    self.hashcode = None

  # Assumes that 'other' is already the correct type
  def __eq__(self, other):
    return (self.args == other.args) and (self.kwargs == other.kwargs)

  def __hash__(self):
    if self.hashcode is None:
      hashable_kwargs = tuple(sorted(self.kwargs.items())) if self.kwargs else ()
      self.hashcode = hash((self.args, hashable_kwargs))
    return self.hashcode

  def __repr__(self):
    args_str = (", ").join([str(i) for i in self.args] + ["%s=%s" % (k, v) for (k, v) in self.kwargs.items()])
    return "%s(%s)" % (self.__class__.__name__, args_str)


class NoResultException(Exception):
  pass


class CacheEntry():

  def __init__(self, cache, args, kwargs):
    self.cache = cache
    self.args = args
    self.kwargs = kwargs

    self.updated_at = None
    self.load_lock = Lock()
    self.get_value()
    if self.cache.auto_populate_s is not None:
      self.start_auto_populate_timer()
 
  def __del__(self):
    self.did_evict()

  def did_evict(self):
    # We explicitly set 'is_evicted' instead of just cancelling the timer in case our auto-populate worker is in the
    # midst of doing work. In that case, the timer is not active, and the call to 'cancel' is meaningless
    if hasattr(self, 'timer'):
      self.is_evicted = True
      self.timer.cancel()

  def start_auto_populate_timer(self):
    # Note: this creates a new thread each time, which is presumably more inefficient...
    self.timer = Timer(self.cache.auto_populate_s, self.auto_populate)
    self.timer.start()

  def auto_populate(self):
    self.cache.get_cache_entry(self.args, self.kwargs)
    if not getattr(self, 'is_evicted', False):
      self.start_auto_populate_timer()

  def is_stale(self):
    if (self.cache.auto_populate_s is not None):
      return (time() - self.updated_at) > self.cache.auto_populate_s
    if (self.cache.ttl_s is None) or (not self.updated_at):
      return False
    return (time() - self.updated_at) > self.cache.ttl_s

  def check_result(self):
    if hasattr(self, 'value'):
      return self.value
    if hasattr(self, 'exception'):
      raise self.exception
    raise NoResultException()

  def get_value(self):
    with suppress(NoResultException):
      return self.check_result()

    self.load_lock.acquire()

    with suppress(NoResultException):
      return self.check_result()

    try:
      self.value = self.cache.fn(*self.args, **self.kwargs)
    except Exception as e:
      self.exception = e
    finally:
      self.load_lock.release()

    self.updated_at = time()
    return self.check_result()


CacheInfo = namedtuple("CacheInfo", [ "ttl_s", "max_size", "cur_size", "hits", "misses" ])


class Cache():

  def __init__(self, ttl_s, max_size, auto_populate_s, fn):
    self.ttl_s = ttl_s
    self.max_size = max_size
    self.auto_populate_s = auto_populate_s
    self.fn = fn
    self.hits = 0
    self.misses = 0
    self._contents = OrderedDict()
    self._update_lock = Lock()

  def _evict_key(self, cache_key):
    cache_entry = self._contents[cache_key]
    del self._contents[cache_key]
    cache_entry.did_evict()

  def evict(self, *args, **kwargs):
    key = CacheKey(args, kwargs)
    self._update_lock.acquire()
    try:
      self._evict_key(key)
      return True
    except:
      return False
    finally:
      self._update_lock.release()

  def _get_cache_entry_hit(self, args, kwargs):
    cache_key = CacheKey(args, kwargs)

    cache_entry = self._contents.get(cache_key)
    if cache_entry and (not cache_entry.is_stale()):
      return cache_entry, True

    self._update_lock.acquire()
    try:
      cache_entry = self._contents.get(cache_key)

      if not cache_entry:
        if (self.max_size is not None) and len(self._contents) >= self.max_size:
          _, evicted_cache_entry = self._contents.popitem(last=False)
          evicted_cache_entry.did_evict()
        new_cache_entry = CacheEntry(self, args, kwargs)
        self._contents[cache_key] = new_cache_entry
        return new_cache_entry, False

      if cache_entry.is_stale():
        self._evict_key(cache_key)
        new_cache_entry = CacheEntry(self, args, kwargs)
        self._contents[cache_key] = new_cache_entry
        return new_cache_entry, False

      return cache_entry, True
        
    finally:
      self._update_lock.release()

  def get_cache_entry(self, args, kwargs):
    cache_entry, is_hit = self._get_cache_entry_hit(args, kwargs)
    if is_hit:
      self.hits += 1
    else:
      self.misses += 1
    return cache_entry

  def info(self):
    return CacheInfo(self.ttl_s, self.max_size, len(self._contents), self.hits, self.misses)


def cache(ttl_s=None, max_size=128, auto_populate_s=None):
  def decorator(f):
    cache_manager = Cache(ttl_s, max_size, auto_populate_s, f)
    @wraps(f)
    def run(*args, **kwargs):
      return cache_manager.get_cache_entry(args, kwargs).get_value()
    setattr(run, 'cache', cache_manager)
    return run
  return decorator




# TESTS

# from collections import defaultdict
# from threading import Thread
# from time import sleep




# test_autopopulate_maxsize

# work_done = defaultdict(lambda: 0)
# @cache(max_size=2, auto_populate_s=1)
# def test(i):
#   work_done[i] += 1
# test(1) # work_done[1] += 1
# test(2) # work_done[2] += 1
# sleep(2.5) # work_done[1] += 2, work_done[2] += 2 (due to auto-population)
# test(1)
# test(2)
# assert work_done[1] == 3
# assert work_done[2] == 3
# test(3) # work_done[3] += 1 (evicts 1 from the cache)
# sleep(1) # work_done[1] += 1, work_done[2] += 1 (due to auto-population)
# assert work_done[1] == 3
# assert work_done[2] == 4
# assert work_done[3] == 2
# test.cache.evict(2)
# test.cache.evict(3)
# sleep(1.5) # work_done should not change due to auto-population
# assert work_done[1] == 3
# assert work_done[2] == 4
# assert work_done[3] == 2




# test_autopopulate_exception
# An exception should cause the auto-population to terminate (we don't cache exceptions)

# work_done = 0
# @cache(auto_populate_s=1)
# def test():
#   global work_done
#   work_done += 1
#   raise Exception()
# try:
#   test()
# except:
#   pass
# assert work_done == 1
# sleep(1.5) # If the auto-populate had been running, we would do more work here
# assert work_done == 1






# test_multithreading_uncached
# When there is no cached value and the function is called more than once while it is still working, the callers should
# share the results

# work_done = 0
# @cache()
# def test():
#   global work_done
#   work_done += 1
#   sleep(0.5)
# threads = [ Thread(target=test) for _ in range(3) ]
# [ i.start() for i in threads ]
# [ i.join() for i in threads ]
# assert work_done == 1



# test_multithreading_uncached_exception

# exceptions = 0
# @cache()
# def test():
#   sleep(0.5)
#   raise Exception()
# def expect_exception():
#   global exceptions
#   try:
#     test()
#   except:
#     exceptions += 1
# threads = [ Thread(target=expect_exception) for _ in range(3) ]
# [ i.start() for i in threads ]
# [ i.join() for i in threads ]
# assert exceptions == 3




# TODO: I have not tested all the nuances with auto-population and multi-threading





# test_multi_thread

# work_done = defaultdict(lambda: 0)
# @cache()
# def test(i):
#   work_done[i] += 1
# def t_1():
#   test(1)
# def t_2():
#   sleep(0.5)
#   test(1)


# try:
#   test()
# except:
#   pass
# assert work_done == 1
# sleep(1.5) # If the auto-populate had been running, we would do more work here
# assert work_done == 1







# import os
# os._exit(0)




# @cache(auto_populate_s=2)
# def test(a):
#   print('test(%s)' % a)
#   sleep(1)
#   return a

# print(test(1))
# sleep(6)
# print(test(1))